;macros.inc
;21-Aug-03

; -------------------------------------------------------------------------
; Macros from various sources
; -------------------------------------------------------------------------

option dotname

; This one comes with Hutch's Masm32 package
AppModel MACRO Processor
  Processor             ;; Processor type
  .model flat, stdcall  ;; 32 bit memory model
  option casemap :none  ;; case sensitive
ENDM

; This one comes with RadASM
m2m macro m1:REQ,m2:REQ
    push &m2&
    pop &m1&
endm

; I'm not sure where this ones came from...
x2m macro m1:REQ,m2:REQ
    push &m1&
    push &m2&
    pop &m1&
    pop &m2&
endm

$invoke MACRO vars:VARARG
     invoke vars
     EXITM <eax>
ENDM

.block      equ <.repeat>
.endblock   equ <.until TRUE>
.forever    equ <.until FALSE>

;From MASM32.HLP
@ArgI MACRO index:REQ, arglist:VARARG
    LOCAL count, retstr
    count = 0
    FOR arg, <arglist>
        count = count + 1
        IF count EQ index
            retstr textequ <arg>
        ENDIF
    ENDM
    EXITM <retstr>
ENDM

;By Biterider
; Macro: JumpOn
; Purpose: Creates a jump table and executes a jump to a label according the content of a register.
; Arguments: Arg1: case register.
; Arg2: Jump labels.
; Returns: Nothing.
; Example: JumpOn eax, @@10, @@20, @@30
JumpOn macro RegName:req, Labels:vararg
local JumpTable, Count, Arg, CaseElse

.data
Count = 0
for Arg, <&Labels>
    ife Count
        JumpTable dd offset Arg
    else
        dd offset Arg
    endif
    Count = Count + 1
endm
.code
cmp RegName, &Count
jae CaseElse
jmp [offset JumpTable + 4 * RegName]
CaseElse:
endm

; Returns IP address as DWORD in network byte order
; Usage: $IPADDR(192,168,0,1)
; by Exagone [Thomas Bleeker]
$IPADDR    MACRO IPP1:REQ,IPP2:REQ,IPP3:REQ,IPP4:REQ
LOCAL    @val
    @val = IPP1 + (IPP2 SHL 8) + (IPP3 SHL 16) + (IPP4 SHL 24)
    EXITM    <&@val>
ENDM

; Returns port number as WORD in network byte order
; Usage: $PORTNR(1234)
; by Exagone [Thomas Bleeker]
$PORTNR    MACRO PORTNUMBER:REQ
LOCAL    @val
    @val = (PORTNUMBER SHR 8) + ((PORTNUMBER AND 0FFh) SHL 8)
    EXITM    <&@val>
ENDM

; -------------------------------------------------------------------------
; Some macros of my own
; -------------------------------------------------------------------------

;needs immediate values as params
;returns 00bbggrr (COLORREF order)
$RGB MACRO red,green,blue
    EXITM %(((blue) SHL 16) OR ((green) SHL 8) OR (red))
ENDM

; Define ASCIIZ string
sz macro arg
    sz&arg& db "&arg&",0
endm

 ; incl macro arg
 ;     include \masm32\include\&arg.inc
 ;     includelib \masm32\lib\&arg&.lib
 ; endm

; Include API library
incl macro var:VARARG
    local count,countmax
    countmax = 0
    for arg, <var>
        countmax = countmax + 1
    endm
    if countmax eq 0
        exitm
    endif
    count = 1
:incl_loop
    incfile textequ @ArgI ( count, <var> )
    @CatStr ( <include \masm32\include\>, %incfile, <.inc> )
    @CatStr ( <includelib \masm32\lib\>, %incfile, <.lib> )
    if count eq countmax
        exitm
    endif
    count = count + 1
    goto incl_loop
endm

; Include snippet
incls macro var:VARARG
    local count,countmax
    countmax = 0
    for arg, <var>
        countmax = countmax + 1
    endm
    if countmax eq 0
        exitm
    endif
    count = 1
:incls_loop
    incfile textequ @ArgI ( count, <var> )
    @CatStr ( <include \masm32\snippets\>, %incfile, <.inc> )
    if count eq countmax
        exitm
    endif
    count = count + 1
    goto incls_loop
endm

;this two were inspired in some macros found in MACROS.ASM
movi macro reg:REQ,imm:REQ
    IF imm EQ 0
        xor reg,reg
    ELSEIF imm EQ 1
        xor reg,reg
        inc reg
    ELSEIF imm EQ -1
        xor reg,reg
        dec reg
    ELSE
        mov reg,&imm&
    ENDIF
endm

reti MACRO imm:REQ
    movi eax,&imm&
    ret
ENDM

;this macros are for switch/case support (similar to C)

; mov eax,uMsg
; .switch eax
; .case WM_COMMAND
;   (...)
;   .break
; .case WM_NOTIFY
;   (...)
;   .break
; .default
;   (...)
; .endswitch

; .switch {register or variable}
.switch macro regname:req
    ifndef SwitchNesting
        SwitchNesting = 1
    else
        SwitchNesting = SwitchNesting + 1
    endif
    @CatStr ( <SwitchRegister_>, %SwitchNesting ) textequ <&regname&>
    @CatStr ( <SwitchFirstCase_>, %SwitchNesting ) = 1
    @CatStr ( <SwitchLastCase_>, %SwitchNesting ) = 0
    @CatStr ( <SwitchNeedDefault_>, %SwitchNesting ) = 0
    .repeat
endm

; .case {value 1}, [value 2], (...), [value n]
.case macro var:vararg
    local value,isfirst,cond

    ifndef SwitchNesting
        .err <.case cannot be outside of a .switch/.endswitch block.>
        exitm
    endif
    if SwitchNesting lt 1
        .err <Bad .switch/.endswitch block.>
        exitm
    endif
    ife @CatStr ( <SwitchLastCase_>, %SwitchNesting )
        ife @CatStr ( <SwitchFirstCase_>, %SwitchNesting )
            .endif
        else
            @CatStr ( <SwitchFirstCase_>, %SwitchNesting ) = 0
        endif
        isfirst = 1
        for value, <var>
            if isfirst ne 0
                cond catstr <( SwitchRegister_>, %SwitchNesting, < == &value& )>
                isfirst = 0
            else
                cond textequ @CatStr ( %cond, < || ( SwitchRegister_>, %SwitchNesting, < == >, <&value&>, < )> )
            endif
        endm
        if isfirst eq 0
            .if cond
        endif
    else
        .err <.case cannot follow .default.>
    endif
endm

; .between {min 1}, {max 1}, [min 2], [max 2], (...), [min n], [max n]
; values must always be given in pairs
.between macro var:vararg
    local value,isfirst,iseven,cond,cond2

    ifndef SwitchNesting
        .err <.between cannot be outside of a .switch/.endswitch block.>
        exitm
    endif
    if SwitchNesting lt 1
        .err <Bad .switch/.endswitch block.>
        exitm
    endif
    ife @CatStr ( <SwitchLastCase_>, %SwitchNesting )
        ife @CatStr ( <SwitchFirstCase_>, %SwitchNesting )
            .endif
        else
            @CatStr ( <SwitchFirstCase_>, %SwitchNesting ) = 0
        endif
        isfirst = 1
        iseven = 1
        for value, <var>
            if isfirst ne 0
                ;( ( reg >= value )
                cond catstr <( ( SwitchRegister_>, %SwitchNesting, < !>= &value& )>
                isfirst = 0
            else
                if iseven eq 1
                    ; || ( ( reg >= value )
                    cond2 catstr <cond>, < || ( ( SwitchRegister_>, %SwitchNesting, < !>= &value& )>
                    cond textequ cond2
                else
                    ; && ( reg <= value ) )
                    cond2 catstr <cond>, < && ( SwitchRegister_>, %SwitchNesting, < !<= &value& ) )>
                    cond textequ cond2
                endif
            endif
            iseven = -iseven
        endm
        if isfirst eq 0
            .if cond
        endif
    else
        .err <.between cannot follow .default.>
    endif
endm

; .caseif {condition}
; The condition must be a text literal to be passed to the .if macro.
; Remember to escape angle brackets with an exclamation point.
.caseif macro cond:req
    ifndef SwitchNesting
        .err <.caseif cannot be outside of a .switch/.endswitch block.>
        exitm
    endif
    if SwitchNesting lt 1
        .err <Bad .switch/.endswitch block.>
        exitm
    endif
    ife @CatStr ( <SwitchLastCase_>, %SwitchNesting )
        ife @CatStr ( <SwitchFirstCase_>, %SwitchNesting )
            .endif
        else
            @CatStr ( <SwitchFirstCase_>, %SwitchNesting ) = 0
        endif
        .if cond
    else
        .err <.caseif cannot follow .default.>
    endif
endm

.default macro
    ifndef SwitchNesting
        .err <.default cannot be outside of a .switch/.endswitch block.>
        exitm
    endif
    if SwitchNesting lt 1
        .err <Bad .switch/.endswitch block.>
        exitm
    endif
    ifndef SwitchCount
        SwitchCount = 0
    endif
    ife @CatStr ( <SwitchLastCase_>, %SwitchNesting )
        ife @CatStr ( <SwitchFirstCase_>, %SwitchNesting )
            .endif
        else
            .err <There should be at least one .case before .default.>
        endif
        @CatStr ( <SwitchLastCase_>, %SwitchNesting ) = 1
    else
        .err <There cannot be more that one .default in a .switch/.endswitch block.>
    endif
    @CatStr ( <SwitchNeedDefault_>, %SwitchNesting ) = 0
    @CatStr ( <SwitchLastCase_>, %SwitchNesting ) = 1
    @CatStr ( <SwitchDefault_>, %SwitchCount, <:> )
    SwitchCount = SwitchCount + 1
endm

.breakdef macro
    ifndef SwitchNesting
        .err <.breakdef cannot be outside of a .switch/.endswitch block.>
        exitm
    endif
    if SwitchNesting lt 1
        .err <Bad .switch/.endswitch block.>
        exitm
    endif
    ifndef SwitchCount
        SwitchCount = 0
    endif
    @CatStr ( <SwitchNeedDefault_>, %SwitchNesting ) = 1
    jmp @CatStr ( <SwitchDefault_>, %SwitchCount )
endm

.endswitch macro
    if SwitchNesting lt 1
        .err <Bad .switch/.endswitch block.>
        exitm
    endif
    ife @CatStr ( <SwitchFirstCase_>, %SwitchNesting )
        ife @CatStr ( <SwitchLastCase_>, %SwitchNesting )
            .endif
        endif
    else
        .err <Empty .switch/.endswitch block?>
    endif
    if @CatStr ( <SwitchNeedDefault_>, %SwitchNesting ) ne 0
        .err <Unmatched .breakdef, could branch to unexpected location!>
    endif
    .until TRUE
    SwitchNesting = SwitchNesting - 1
    if SwitchNesting lt 0
        .err <Bad .switch/.endswitch block.>
    endif
endm
